"""Functions for data validation."""

import base64
import binascii
import hashlib
import math
import string
from collections.abc import Callable, Iterable
from functools import cache
from typing import Any, overload

from algosdk.encoding import is_valid_address
from babel import Locale, UnknownLocaleError
from pydantic import TypeAdapter
from pydantic_core import Url

from algobase.utils.read import read_ipfs_gateways, read_mime_types


def is_valid(func: Callable[..., Any], *args: Any, **kwargs: Any) -> bool:
    """Checks if a function call is valid.

    The other functions in this module raise errors when the input is not valid.
    This is a convenience function to check if a function call is valid without raising an error.

    Args:
        func (Callable[..., Any]): The function to call.
        *args: Variable length argument list.
        **kwargs: Arbitrary keyword arguments.

    Returns:
        bool: True if the function call doesn't raise a ValueError, else False.
    """
    try:
        func(*args, **kwargs)
        return True
    except ValueError:
        return False


def validate_address(value: str) -> str:
    """Checks that the value is a valid Algorand address.

    Args:
        value (str): The value to check.

    Raises:
        ValueError: If the value is not a valid Algorand address.

    Returns:
        str: The value passed in.
    """
    if not is_valid_address(value):
        raise ValueError(f"'{value}' is not a valid Algorand address.")
    return value


@overload  # pragma: no cover
def validate_encoded_length(value: str, max_length: int) -> str:
    ...


@overload  # pragma: no cover
def validate_encoded_length(value: Url, max_length: int) -> Url:
    ...


def validate_encoded_length(value: str | Url, max_length: int) -> str | Url:
    """Checks that the value is not longer than `max_length` when encoded in UTF-8.

    Args:
        value (str | Url): The value to check.
        max_length (int): The maximum length of the value when encoded in UTF-8.

    Raises:
        ValueError: If the value is longer than `max_length` when encoded in UTF-8.

    Returns:
        str | Url: The value passed in.
    """
    url = value if isinstance(value, str) else value.unicode_string()
    if len(url.encode("utf-8")) > max_length:
        raise ValueError(f"'{value}' is > {max_length} bytes when encoded in UTF-8.")
    return value


@cache
def validate_not_ipfs_gateway(url: str) -> str:
    """Checks that the URL host is not a known public IPFS gateway.

    Args:
        url (str): The URL to check.

    Raises:
        ValueError: If the URL host is a known public IPFS gateway.

    Returns:
        str: The URL passed in.
    """
    gateways = read_ipfs_gateways()
    if any(Url(gateway).host == Url(url).host for gateway in gateways):
        raise ValueError(f"'{Url(url).host}' is an IPFS gateway.")
    return url


def validate_base64(value: str) -> str:
    """Checks that the value is a valid base64 string.

    Args:
        value (str): The value to check.

    Raises:
        ValueError: If the value is not a valid base64 string.

    Returns:
        str: The value passed in.
    """
    try:
        base64.b64decode(value, validate=True)
    except binascii.Error:
        raise ValueError(f"'{value}' is not valid base64.")
    return value


def validate_sri(value: str) -> str:
    """Checks that the value is a valid W3C Subresource Integrity (SRI) value.

    Args:
        value (str): The value to check.

    Raises:
        ValueError: If the value is not a valid SRI.

    Returns:
        str: The value passed in.
    """
    supported_algorithms = {"sha256", "sha384", "sha512"}
    hash_algorithm = next(
        (x for x in supported_algorithms if value.startswith(f"{x}-")), None
    )
    if hash_algorithm is None:
        raise ValueError(
            f"'{value}' is not a valid SRI. String must start with 'sha256-', 'sha384-', or 'sha512-'."
        )
    hasher = hashlib.new(hash_algorithm)
    hash_digest = value.removeprefix(f"{hash_algorithm}-")
    try:
        validate_base64(hash_digest)
    except ValueError as e:
        raise ValueError(f"'{value}' is not a valid SRI. Hash digest {e}")
    if len(base64.b64decode(hash_digest)) != hasher.digest_size:
        raise ValueError(
            f"'{value}' is not a valid SRI. Expected {hasher.digest_size} byte hash digest, got {len(hash_digest)} bytes."
        )
    return value


def validate_arc3_sri(value: str) -> str:
    """Checks that the value is a valid SHA-256 Subresource Integrity (SRI) value.

    Args:
        value (str): The value to check.

    Raises:
        ValueError: If the value is not a valid SRI.

    Returns:
        str: The value passed in.
    """
    if not value.startswith("sha256-"):
        raise ValueError(
            f"'{value}' is not a valid ARC-3 SRI. String must start with 'sha256-'."
        )
    return validate_sri(value)


@cache
def validate_mime_type(value: str, primary_type: str | None = None) -> str:
    """Checks that the value is a valid MIME type.

    If `primary_type` is not `None`, then the value must be a valid MIME type with the specified primary type.
    E.g. if `primary_type` is 'image', then the value must be a valid MIME type starting with 'image/'.

    Args:
        value (str): The value to check.
        primary_type (str | None, optional): The primary type of the MIME type. Defaults to None.

    Raises:
        ValueError: If the value is not a valid MIME type.

    Returns:
        str: The value passed in.
    """
    if value not in read_mime_types():
        raise ValueError(f"'{value}' is not a valid MIME type.")
    if primary_type is not None and not value.startswith(f"{primary_type}/"):
        raise ValueError(f"'{value}' is not a valid {primary_type} MIME type.")
    return value


def validate_hex(value: str) -> str:
    """Checks that the value is a valid hexadecimal string.

    Args:
        value (str): The value to check.

    Raises:
        ValueError: If the value is not a valid hexadecimal string.

    Returns:
        str: The value passed in.
    """
    if not all(x in string.hexdigits for x in value):
        raise ValueError(f"'{value}' is not a valid hex string.")
    return value


def validate_locale(value: str) -> str:
    """Checks that the value is a valid Unicode CLDR locale.

    Args:
        value (str): The value to check.

    Raises:
        ValueError: If the value is not a valid locale identifier.
        UnknownLocaleError: If the value is not a valid Unicode CLDR locale.

    Returns:
        str: The value passed in.
    """
    try:
        Locale.parse(value)
    except ValueError as e:
        raise ValueError(f"'{value}' is not a valid locale identifier: {e}")
    except UnknownLocaleError:
        raise ValueError(f"'{value}' is not a valid Unicode CLDR locale.")
    return value


def validate_contains_substring(value: str | Url, substring: str) -> str | Url:
    """Checks that the value contains the substring.

    Args:
        value (str | Url): The value to check.
        substring (str): The substring to check for.

    Raises:
        ValueError: If the value does not contain the substring.

    Returns:
        str | Url: The value passed in.
    """
    value_string = value if isinstance(value, str) else value.unicode_string()
    if substring not in value_string:
        raise ValueError(f"'{value_string}' does not contain subtring '{substring}'.")
    return value


def validate_not_in(iterable: Iterable[str], element: str) -> Iterable[str]:
    """Checks that the element is not in the iterable.

    Args:
        element (str): The element to check for.
        iterable (Iterable): The iterable to check.

    Raises:
        ValueError: If the element is in the iterable.

    Returns:
        Iterable: The iterable passed in.
    """
    if element in iterable:
        raise ValueError(f"'{element}' is in {iterable}.")
    return iterable


def validate_is_power_of_10(n: int) -> int:
    """Checks that the value is a power of 10.

    Args:
        n (int): The value to check.

    Raises:
        ValueError: If the value is not a power of 10.

    Returns:
        int: The value passed in.
    """
    if not (n > 0 and math.log10(n).is_integer()):
        raise ValueError(f"{n} is not a power of 10.")
    return n


def validate_type_compatibility(value: str, _type: type) -> str:
    """Checks that the value is compatible with the annotated type.

    Args:
        value (str): The value to check.
        _type (Type): The type to validate against.

    Raises:
        ValidationError: If the value is not compatible with the type.

    Returns:
        str: The value passed in.
    """
    TypeAdapter(_type).validate_python(value)
    return value



"""Configuration settings for the algobase."""

from dynaconf import Dynaconf, Validator

from algobase.choices import AlgorandNetwork

settings = Dynaconf(
    envvar_prefix="AB",  # Abbreviation for 'algobase'
    settings_files=["settings.toml", ".secrets.toml"],
    environments=True,
    load_dotenv=True,
    env_switcher="AB_ENV",
    validators=[
        Validator(
            "ALGORAND_NETWORK",
            when=Validator("NAME", eq="") | Validator("NAME", eq="dev"),
            default=AlgorandNetwork.LOCALNET,
        ),
        Validator(
            "ALGORAND_NETWORK",
            when=Validator("NAME", eq="test"),
            default=AlgorandNetwork.TESTNET,
        ),
        Validator(
            "ALGORAND_NETWORK",
            when=Validator("NAME", eq="prod"),
            default=AlgorandNetwork.MAINNET,
        ),
        Validator(
            "ALGORAND_NETWORK",
            condition=lambda x: x in AlgorandNetwork._value2member_map_,
            messages={
                "condition": "{name} configuration invalid in {env} environment. Found '{value}' but expected 'localnet', 'testnet', or 'mainnet'.",
            },
        ),
        Validator("ALGORAND_NETWORK", cast=AlgorandNetwork),
        Validator(
            "ALGOD_TOKEN",
            when=Validator("ALGORAND_NETWORK", eq="localnet"),
            default="a" * 64,
        ),
        Validator("ALGOD_TOKEN", cast=str),
    ],
)



"""Enums and enum type aliases for algobase."""

from enum import StrEnum, auto
from typing import Literal, TypeAlias


class Arc(StrEnum):
    """An enumeration of Algorand ARC standards that are supported in algobase."""

    ARC3 = auto()


ArcChoice: TypeAlias = Literal[Arc.ARC3]


class AsaType(StrEnum):
    """An enumeration of Algorand Standard Asset (ASA) types."""

    FUNGIBLE = auto()
    NON_FUNGIBLE_PURE = auto()
    NON_FUNGIBLE_FRACTIONAL = auto()


AsaTypeChoice: TypeAlias = Literal[
    AsaType.FUNGIBLE, AsaType.NON_FUNGIBLE_PURE, AsaType.NON_FUNGIBLE_FRACTIONAL
]


class IpfsProvider(StrEnum):
    """An enumeration of IPFS providers."""

    NFT_STORAGE = auto()


IpfsProviderChoice: TypeAlias = Literal[IpfsProvider.NFT_STORAGE]


class IpfsPinStatus(StrEnum):
    """An enumeration of IPFS pin statuses."""

    QUEUED = auto()
    PINNING = auto()
    PINNED = auto()
    FAILED = auto()


IpfsPinStatusChoice: TypeAlias = Literal[
    IpfsPinStatus.QUEUED,
    IpfsPinStatus.PINNING,
    IpfsPinStatus.PINNED,
    IpfsPinStatus.FAILED,
]


class AlgorandNetwork(StrEnum):
    """An enumeration of Algorand networks."""

    LOCALNET = auto()
    TESTNET = auto()
    MAINNET = auto()


AlgorandNetworkChoice: TypeAlias = Literal[
    AlgorandNetwork.LOCALNET, AlgorandNetwork.TESTNET, AlgorandNetwork.MAINNET
]


class AlgorandApi(StrEnum):
    """An enumeration of Algorand APIs."""

    ALGOD = auto()
    INDEXER = auto()
    KMD = auto()


AlgorandApiChoice: TypeAlias = Literal[
    AlgorandApi.ALGOD, AlgorandApi.INDEXER, AlgorandApi.KMD
]


class AlgorandApiProvider(StrEnum):
    """An enumeration of Algorand API providers."""

    LOCALHOST = auto()
    CUSTOM = auto()
    ALGONODE = auto()


AlgorandApiProviderChoice: TypeAlias = Literal[
    AlgorandApiProvider.LOCALHOST,
    AlgorandApiProvider.CUSTOM,
    AlgorandApiProvider.ALGONODE,
]



"""Functions for type casting."""

from collections.abc import Callable
from typing import TypeVar

a = TypeVar("a")
b = TypeVar("b")


def maybe_bind(x: a | None, f: Callable[[a], b | None]) -> b | None:
    """Return the result of applying the function to the value if the value is not None, otherwise return None.

    Args:
        x (a | None): The value to apply the function to.
        f (Callable[[a], b  |  None]): The function to apply to the value.

    Returns:
        _type_: _description_
    """
    return f(x) if x is not None else None



"""A Pydantic model for Algorand Standard Asset parameters."""

from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
)

from algobase.types.annotated import (
    AlgorandAddress,
    AlgorandHash,
    AsaAssetName,
    AsaDecimals,
    AsaUnitName,
    AsaUrl,
    Uint64,
)


class AssetParams(BaseModel):
    """A Pydantic model for Algorand Standard Assets (ASAs)."""

    model_config = ConfigDict(frozen=True)

    total: Uint64 = Field(
        description="The total number of base units of the asset to create."
    )
    decimals: AsaDecimals = Field(
        default=0,
        description="The number of digits to use after the decimal point when displaying the asset. If 0, the asset is not divisible. If 1, the base unit of the asset is in tenths. Must be between 0 and 19, inclusive.",
    )
    default_frozen: bool = Field(
        default=False,
        description="Whether slots for this asset in user accounts are frozen by default.",
    )
    unit_name: AsaUnitName | None = Field(
        default=None,
        description="The name of a unit of this asset. Max size is 8 bytes. Example: 'USDT'.",
    )
    asset_name: AsaAssetName | None = Field(
        default=None,
        description="The name of the asset. Max size is 32 bytes. Example: 'Tether'.",
    )
    url: AsaUrl | None = Field(
        default=None,
        description="Specifies a URL where more information about the asset can be retrieved. Max size is 96 bytes.",
    )
    metadata_hash: AlgorandHash | None = Field(
        default=None,
        description="This field is intended to be a 32-byte hash of some metadata that is relevant to your asset and/or asset holders.",
    )
    manager: AlgorandAddress | None = Field(
        default=None,
        description="The address of the account that can manage the configuration of the asset and destroy it.",
    )
    reserve: AlgorandAddress | None = Field(
        default=None,
        description="The address of the account that holds the reserve (non-minted) units of the asset.",
    )
    freeze: AlgorandAddress | None = Field(
        default=None,
        description="The address of the account used to freeze holdings of this asset. If empty, freezing is not permitted.",
    )
    clawback: AlgorandAddress | None = Field(
        default=None,
        description="The address of the account that can clawback holdings of this asset. If empty, clawback is not permitted.",
    )



"""Pydantic models for Algorand ARC-3 metadata.

Reference: https://github.com/algorandfoundation/ARCs/blob/main/ARCs/arc-0003.md
"""

from typing import Literal

from pydantic import BaseModel, ConfigDict, Field

from algobase.choices import Arc
from algobase.types.annotated import (
    Arc3Color,
    Arc3LocalizedUrl,
    Arc3Sri,
    Arc3Url,
    Arc16Traits,
    AsaDecimals,
    Base64Str,
    ImageMimeType,
    MimeType,
    UnicodeLocale,
)


class Arc3Localization(BaseModel):
    """A Pydantic model for Algorand ARC-3 localization."""

    model_config = ConfigDict(frozen=True)

    uri: Arc3LocalizedUrl = Field(
        description="The URI pattern to fetch localized data from. This URI should contain the substring `{locale}` which will be replaced with the appropriate locale value before sending the request."
    )
    default: UnicodeLocale = Field(
        description="The locale of the default data within the base JSON."
    )
    locales: list[UnicodeLocale] = Field(
        description="The list of locales for which data is available. These locales should conform to those defined in the Unicode Common UnicodeLocale Data Repository (http://cldr.unicode.org/)."
    )
    integrity: dict[UnicodeLocale, Arc3Sri] | None = Field(
        default=None,
        description="The SHA-256 digests of the localized JSON files (except the default one). The field name is the locale. The field value is a single SHA-256 integrity metadata as defined in the W3C subresource integrity specification (https://w3c.github.io/webappsec-subresource-integrity).",
    )


class Arc3Properties(BaseModel):
    """A Pydantic model for Algorand ARC-3 properties.

    If the `traits` property is present, it must comply with ARC-16: https://github.com/algorandfoundation/ARCs/blob/main/ARCs/arc-0016.md
    """

    model_config = ConfigDict(frozen=True, extra="allow")

    # Struggling to get recursive type definition working here.
    # Have defined `Arc3NonTraitProperties` in algobase/types/annotated.py
    # but it doesn't work as an annotation for __pydantic_extra__.
    __pydantic_extra__: dict[str, str | int | float | dict | list]  # type: ignore

    traits: Arc16Traits | None = Field(
        default=None,
        description="Traits (attributes) that can be used to calculate things like rarity. Values may be strings or numbers.",
    )


class Arc3Metadata(BaseModel):
    """A Pydantic model for Algorand ARC-3 metadata."""

    model_config = ConfigDict(frozen=True)

    arc: Literal[Arc.ARC3] = Field(
        default=Arc.ARC3,
        description="Name of the Algorand ARC standard that the NFT metadata adheres to.",
        exclude=True,
    )

    @property
    def json_str(self) -> str:
        """Returns the model JSON as a string."""
        return self.model_dump_json(exclude_none=True, indent=4)

    @property
    def json_bytes(self, encoding: Literal["utf-8"] = "utf-8") -> bytes:
        """Returns the model JSON encoded as bytes.

        Currently only officially supports UTF-8 encoding.
        """
        return self.json_str.encode(encoding)

    name: str | None = Field(
        default=None, description="Identifies the asset to which this token represents."
    )
    decimals: AsaDecimals | None = Field(
        default=None,
        description="The number of decimal places that the token amount should display - e.g. 18, means to divide the token amount by 1000000000000000000 to get its user representation.",
    )
    description: str | None = Field(
        default=None, description="Describes the asset to which this token represents."
    )
    image: Arc3Url | None = Field(
        default=None,
        description="A URI pointing to a file with MIME type image/* representing the asset to which this token represents. Consider making any images at a width between 320 and 1080 pixels and aspect ratio between 1.91:1 and 4:5 inclusive.",
    )
    image_integrity: Arc3Sri | None = Field(
        default=None,
        description="The SHA-256 digest of the file pointed by the URI image. The field value is a single SHA-256 integrity metadata as defined in the W3C subresource integrity specification (https://w3c.github.io/webappsec-subresource-integrity).",
    )
    image_mimetype: ImageMimeType | None = Field(
        default=None,
        description="The MIME type of the file pointed by the URI image. MUST be of the form 'image/*'.",
    )
    background_color: Arc3Color | None = Field(
        default=None,
        description="Background color do display the asset. MUST be a six-character hexadecimal without a pre-pended #.",
    )
    external_url: Arc3Url | None = Field(
        default=None,
        description="A URI pointing to an external website presenting the asset.",
    )
    external_url_integrity: Arc3Sri | None = Field(
        default=None,
        description="The SHA-256 digest of the file pointed by the URI external_url. The field value is a single SHA-256 integrity metadata as defined in the W3C subresource integrity specification (https://w3c.github.io/webappsec-subresource-integrity).",
    )
    external_url_mimetype: Literal["text/html"] | None = Field(
        default=None,
        description="The MIME type of the file pointed by the URI external_url. It is expected to be 'text/html' in almost all cases.",
    )
    animation_url: Arc3Url | None = Field(
        default=None,
        description="A URI pointing to a multi-media file representing the asset.",
    )
    animation_url_integrity: Arc3Sri | None = Field(
        default=None,
        description="The SHA-256 digest of the file pointed by the URI external_url. The field value is a single SHA-256 integrity metadata as defined in the W3C subresource integrity specification (https://w3c.github.io/webappsec-subresource-integrity).",
    )
    animation_url_mimetype: MimeType | None = Field(
        default=None,
        description="The MIME type of the file pointed by the URI animation_url. If the MIME type is not specified, clients MAY guess the MIME type from the file extension or MAY decide not to display the asset at all. It is STRONGLY RECOMMENDED to include the MIME type.",
    )
    properties: Arc3Properties | None = Field(
        default=None,
        description="Arbitrary properties (also called attributes). Values may be strings, numbers, object or arrays.",
    )
    extra_metadata: Base64Str | None = Field(
        default=None,
        description="Extra metadata in base64. If the field is specified (even if it is an empty string) the asset metadata (am) of the ASA is computed differently than if it is not specified.",
    )
    localization: Arc3Localization | None = Field(
        default=None,
        description="A sub-object that may be used to provide localized values for fields that need it.",
    )



"""IPFS client for nft.storage."""

from dataclasses import dataclass

import httpx

from algobase.choices import (
    IpfsPinStatus,
    IpfsPinStatusChoice,
    IpfsProvider,
    IpfsProviderChoice,
)
from algobase.config import settings
from algobase.functional import maybe_bind
from algobase.ipfs.client_base import IpfsClient


@dataclass
class NftStorage(IpfsClient):
    """IPFS client for nft.storage.

    Requires the `NFT_STORAGE_API_KEY` environment variable to be set.
    """

    @property
    def ipfs_provider_name(self) -> IpfsProviderChoice:
        """The name of the IPFS provider."""
        return IpfsProvider.NFT_STORAGE

    @property
    def api_version(self) -> str:
        """The version of the IPFS provider's API."""
        return "1.0"

    @property
    def base_url(self) -> httpx.URL:
        """The base URL of the IPFS provider's API."""
        return httpx.URL("https://api.nft.storage")

    @property
    def is_api_key_required(self) -> bool:
        """Whether the IPFS provider requires an API key."""
        return True

    @property
    def api_key(self) -> str | None:
        """The API key."""
        return maybe_bind(settings.get("nft_storage_api_key"), str)

    @property
    def headers(self) -> dict[str, str]:
        """The headers to use for the HTTP requests."""
        return {"Authorization": f"Bearer {self.api_key}"}

    def store_json(self, json: str) -> str:
        """Stores JSON data in IPFS.

        Args:
            json (str): The JSON to store.

        Returns:
            str: The IPFS CID of the stored data.
        """
        with httpx.Client() as client:
            response = client.post(
                url=self.base_url.join("upload"),
                json=json,
                headers=self.headers,
                timeout=10.0,
            )
            data = response.json()
            if response.status_code == httpx.codes.OK:
                if (
                    data.get("ok") is True
                    and (cid := data.get("value").get("cid")) is not None
                ):
                    return str(cid)
                else:
                    raise httpx.HTTPError(
                        f"HTTP Exception for {response.request.url}: Failed to store JSON in IPFS using {self.ipfs_provider_name}."
                    )
            else:
                raise httpx.HTTPError(
                    f"HTTP Exception for {response.request.url}: {response.status_code} {data.get('error').get('message')}"
                )

    def fetch_pin_status(self, cid: str) -> IpfsPinStatusChoice:
        """Returns the pinning status of a file, by CID.

        Args:
            cid (str): The CID of the file to check.

        Returns:
            IpfsPinStatusChoice: The pin status of the CID.
        """
        with httpx.Client() as client:
            response = client.get(
                url=self.base_url.join(f"check/{cid}"),
                headers=self.headers,
                timeout=10.0,
            )
            data = response.json()
            if response.status_code == httpx.codes.OK:
                pin_status = data.get("value").get("pin").get("status")
                if (
                    data.get("ok") is True
                    and pin_status is not None
                    and hasattr(IpfsPinStatus, str(pin_status).upper())
                ):
                    return IpfsPinStatus(pin_status)
                else:
                    raise httpx.HTTPError(
                        f"HTTP Exception for {response.request.url}: {pin_status} is not a valid pin status."
                    )
            else:
                raise httpx.HTTPError(
                    f"HTTP Exception for {response.request.url}: {response.status_code} {data.get('error').get('message')}"
                )



"""Utility functions for hashing data."""

import hashlib


def sha256(data: bytes) -> bytes:
    """Returns a SHA-256 hash digest of the input data.

    Args:
        data (bytes): The data to hash.

    Returns:
        bytes: The hash digest.
    """
    return hashlib.sha256(data).digest()


def sha512_256(data: bytes) -> bytes:
    """Returns a SHA-512/256 hash digest of the input data.

    Args:
        data (bytes): The data to hash.

    Returns:
        bytes: The hash digest.
    """
    return hashlib.new("sha512_256", data).digest()



"""Functions for reading and caching reference data files."""

import mimetypes
import tomllib


def read_ipfs_gateways() -> list[str]:
    """Read IPFS gateways from the reference data file.

    Returns:
        list[str]: The list of IPFS gateways.
    """
    with open("algobase/data/ipfs.toml", "rb") as f:
        data = tomllib.load(f)
    return list(data["ipfs_gateways"])


def read_mime_types() -> list[str]:
    """Read MIME types from the reference data file.

    Returns:
        list[str]: The list of MIME types.
    """
    mimetypes.init()
    return list(mimetypes.types_map.values())



"""Algorand API providers."""

from abc import ABC, abstractmethod
from dataclasses import dataclass
from functools import cached_property
from typing import Literal

import httpx

from algobase.choices import (
    AlgorandApiProvider,
    AlgorandApiProviderChoice,
)
from algobase.config import settings


class AlgorandProvider(ABC):
    """Abstract base class for Algorand API providers."""

    @property
    @abstractmethod
    def provider_name(self) -> AlgorandApiProviderChoice:
        """The name of the Algorand API provider."""
        ...  # pragma: no cover

    @property
    @abstractmethod
    def api_version(self) -> str:
        """The version of the API."""
        ...  # pragma: no cover

    @property
    @abstractmethod
    def credential(self) -> str:
        """The API credential (key or token)."""
        ...  # pragma: no cover

    @property
    @abstractmethod
    def base_url(self) -> httpx.URL:
        """The base URL of the API."""
        ...  # pragma: no cover

    @property
    @abstractmethod
    def algod_url(self) -> httpx.URL:
        """The Algod API base URL."""
        ...  # pragma: no cover

    @property
    @abstractmethod
    def indexer_url(self) -> httpx.URL:
        """The Indexer API base URL."""
        ...  # pragma: no cover

    @property
    @abstractmethod
    def kmd_url(self) -> httpx.URL:
        """The KMD API base URL."""
        ...  # pragma: no cover


@dataclass
class Localhost(AlgorandProvider):
    """Localhost provider."""

    @property
    def provider_name(self) -> Literal[AlgorandApiProvider.LOCALHOST]:
        """The name of the Algorand API provider."""
        return AlgorandApiProvider.LOCALHOST

    @property
    def api_version(self) -> str:
        """The version of the API."""
        return "0"

    @property
    def credential(self) -> str:
        """The API credential (key or token)."""
        return str(settings.algod_token)

    @property
    def base_url(self) -> httpx.URL:
        """The base URL of the API."""
        return httpx.URL("http://localhost")

    @cached_property
    def algod_url(self) -> httpx.URL:
        """The Algod API base URL."""
        return self.base_url.copy_with(port=4001)

    @cached_property
    def indexer_url(self) -> httpx.URL:
        """The Indexer API base URL."""
        return self.base_url.copy_with(port=8980)

    @cached_property
    def kmd_url(self) -> httpx.URL:
        """The KMD API base URL."""
        return self.base_url.copy_with(port=4002)



"""Abstract base class for IPFS clients."""

from abc import ABC, abstractmethod

import httpx

from algobase.choices import IpfsPinStatusChoice, IpfsProviderChoice


class IpfsClient(ABC):
    """Abstract base class for IPFS clients."""

    def __post_init__(self) -> None:
        """If an API key is required, check that it is present."""
        if self.is_api_key_required:
            self.check_api_key_is_present()

    @property
    @abstractmethod
    def ipfs_provider_name(self) -> IpfsProviderChoice:
        """The name of the IPFS provider."""
        ...  # pragma: no cover

    @property
    @abstractmethod
    def api_version(self) -> str:
        """The version of the IPFS provider's API."""
        ...  # pragma: no cover

    @property
    @abstractmethod
    def base_url(self) -> httpx.URL:
        """The base URL of the IPFS provider's API."""
        ...  # pragma: no cover

    @property
    @abstractmethod
    def is_api_key_required(self) -> bool:
        """Whether the IPFS provider requires an API key."""
        ...  # pragma: no cover

    @property
    @abstractmethod
    def api_key(self) -> str | None:
        """The API key."""
        ...  # pragma: no cover

    def check_api_key_is_present(self) -> None:
        """Checks that the IPFS provider's API key is present."""
        if self.is_api_key_required and self.api_key is None:
            raise ValueError(
                f"API key for {self.ipfs_provider_name} must be defined in .env file."
            )

    @abstractmethod
    def store_json(self, json: str) -> str:
        """Stores JSON data in IPFS.

        Args:
            json (str): The JSON to store.

        Returns:
            str: The IPFS CID of the stored data.
        """
        ...  # pragma: no cover

    @abstractmethod
    def fetch_pin_status(self, cid: str) -> IpfsPinStatusChoice:
        """Returns the pinning status of a file, by CID.

        Args:
            cid (str): The CID of the file to check.

        Returns:
            IpfsPinStatusChoice: The pin status of the CID.
        """
        ...  # pragma: no cover



"""Functions for working with URLs."""

from urllib.parse import quote, urlparse


def decode_url_braces(url: str) -> str:
    """Decodes curly braces in a URL string.

    This allows for arbitrary parameters to be passed in URL strings, as specified in some Algorand standards.
    For example, ARC-3 asset URLs may contain the string '{id}', which clients must replace with the asset ID in decimal form.

    Args:
        url (str): The URL string to decode.

    Returns:
        str: The decoded URL string.
    """
    parsed_url = urlparse(url)
    decoded_path = parsed_url.path.replace(quote("{"), "{").replace(quote("}"), "}")
    decoded_url = parsed_url._replace(path=decoded_path).geturl()
    return decoded_url



"""Pydantic models for Algorand Standard Assets (ASAs)."""

import math
import warnings
from base64 import b64decode
from difflib import SequenceMatcher

from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    computed_field,
    model_validator,
)

from algobase.choices import AsaType, AsaTypeChoice
from algobase.models.arc3 import Arc3Metadata
from algobase.models.asset_params import AssetParams
from algobase.types.annotated import AlgorandHash, AsaAssetName
from algobase.utils.hash import sha256, sha512_256
from algobase.utils.validate import is_valid, validate_type_compatibility


class Asa(BaseModel):
    """A Pydantic model for Algorand Standard Assets (ASAs)."""

    model_config = ConfigDict(frozen=True)

    asa_type: AsaTypeChoice | None = Field(
        default=None, description="The type of the ASA."
    )
    asset_params: AssetParams = Field(description="AssetParams Pydantic model.")
    metadata: Arc3Metadata | None = Field(
        default=None, description="Arc3Metadata Pydantic model.", discriminator="arc"
    )

    @property
    def derived_asa_type(self) -> AsaTypeChoice:
        """The derived type of the ASA."""
        match self.asset_params:
            case asa if asa.total == 1 and asa.decimals == 0:
                return AsaType.NON_FUNGIBLE_PURE
            # Means the total supply is 1
            case asa if asa.decimals == math.log10(asa.total):
                return AsaType.NON_FUNGIBLE_FRACTIONAL
            case _:
                return AsaType.FUNGIBLE

    @computed_field  # type: ignore[misc]
    @property
    def metadata_hash(self) -> AlgorandHash | None:
        """The hash of the JSON metadata."""
        # Currently only ARC3 is supported
        if isinstance(self.metadata, Arc3Metadata):
            if self.metadata.extra_metadata is None:
                return sha256(self.metadata.json_bytes)
            else:
                # am = SHA-512/256("arc0003/am" || SHA-512/256("arc0003/amj" || content of JSON Metadata file) || e)
                base_hash = sha512_256(b"arc0003/amj" + self.metadata.json_bytes)
                extra_metadata_bytes = b64decode(self.metadata.extra_metadata)
                return sha512_256(b"arc0003/am" + base_hash + extra_metadata_bytes)
        return None

    @model_validator(mode="after")
    def check_asa_type_constraints(self) -> "Asa":
        """Validate the ASA type against the relevant constraints."""
        if self.asa_type is not None and self.asa_type != self.derived_asa_type:
            match self.asa_type:
                case AsaType.NON_FUNGIBLE_PURE:
                    raise ValueError(
                        "Total number of units must be 1 and number of digits after the decimal point must be 0 for a pure NFT."
                    )
                case AsaType.NON_FUNGIBLE_FRACTIONAL:
                    raise ValueError(
                        "Number of digits after the decimal point must be equal to the logarithm in base 10 of total number of units. In other words, the total supply of the ASA must be exactly 1."
                    )
                case AsaType.FUNGIBLE:
                    raise ValueError(
                        "Total supply of the ASA must be greater than 1, for a fungible asset."
                    )
        return self

    @model_validator(mode="after")
    def check_arc_constraints(self) -> "Asa":
        """Validate fields against ARC constraints, if applicable."""
        # Currently only ARC3 is supported
        if isinstance(self.metadata, Arc3Metadata):
            self.check_arc3_constraints()
        return self

    def check_arc3_decimals(self) -> "Asa":
        """Raise an error if the decimals in the asset parameters doesn't match the deimals in the metadata."""
        if (
            self.metadata is not None
            and self.metadata.decimals is not None
            and self.metadata.decimals != self.asset_params.decimals
        ):
            raise ValueError(
                f"Decimals in the asset parameters ({self.asset_params.decimals}) must match the decimals in the metadata ({self.metadata.decimals})."
            )
        return self

    def check_arc3_unit_name(self) -> "Asa":
        """Raise a warning if the metadata 'name' property is not related to the asset unit name.

        Uses the difflib `SequenceMatcher` for string similarity.
        """
        if (
            self.asset_params.unit_name is not None
            and self.metadata is not None
            and self.metadata.name is not None
            and SequenceMatcher(
                None, self.asset_params.unit_name.lower(), self.metadata.name.lower()
            ).ratio()
            < 0.5
        ):
            warnings.warn(
                UserWarning(
                    "Asset unit name should be related to the name in the ARC-3 JSON metadata."
                )
            )
        return self

    def check_arc3_asset_url(self) -> "Asa":
        """Checks that the asset URL is valid for ARC-3 ASAs."""
        if self.asset_params.url is None:
            raise ValueError("Asset URL must not be `None`.")
        elif not self.asset_params.url.endswith("#arc3"):
            raise ValueError(
                f"Asset URL must end with '#arc3' if asset name is '{self.asset_params.asset_name}'."
            )
        return self

    def check_arc3_constraints(self) -> "Asa":
        """Validate fields against ARC constraints, if applicable.

        Raises warnings for values/formats that are allowed but not recommended in ARC specs.
        Raises errors for values/formats that are not allowed in ARC specs.
        """
        # Currently only ARC3 is supported
        if isinstance(self.metadata, Arc3Metadata):
            self.check_arc3_unit_name()
            self.check_arc3_decimals()

            # Asset name constraints
            match self.asset_params.asset_name:
                case None:
                    raise ValueError("Asset name must not be `None` for ARC-3 ASAs.")
                case "arc3":
                    warnings.warn(
                        UserWarning(
                            "Asset name 'arc3' is not recommended for ARC-3 ASAs."
                        )
                    )
                case x if x.endswith("@arc3"):
                    warnings.warn(
                        UserWarning(
                            "Asset name format <name>@arc3 is not recommended for ARC-3 ASAs."
                        )
                    )
                # Constraints on combination of asset name and metadata name
                case _:
                    match self.metadata.name:
                        case None:
                            raise ValueError(
                                f"Metadata name must not be `None` if asset name is '{self.asset_params.asset_name}'."
                            )
                        case x if x != self.asset_params.asset_name:
                            if is_valid(
                                validate_type_compatibility,
                                self.metadata.name,
                                AsaAssetName,
                            ):
                                raise ValueError(
                                    f"Asset name '{self.asset_params.asset_name}' must match the metadata name '{x}'."
                                )
                            elif not self.metadata.name.startswith(
                                self.asset_params.asset_name
                            ):
                                raise ValueError(
                                    f"Asset name must be a shortened version of the metadata name '{self.metadata.name}'."
                                )
                    # Constraints on combination of asset name and asset URL
                    self.check_arc3_asset_url()
        return self



"""Algorand base client."""

from dataclasses import dataclass, field

from algosdk.v2client.algod import AlgodClient

from algobase.algorand.providers import AlgorandProvider, Localhost
from algobase.choices import AlgorandNetworkChoice
from algobase.config import settings


@dataclass
class AlgorandClient:
    """Algorand base client."""

    network: AlgorandNetworkChoice = settings.algorand_network

    provider: AlgorandProvider = field(default_factory=Localhost)

    def __post_init__(self) -> None:
        """Instantiate algosdk clients."""
        self.algod = AlgodClient(
            algod_token=self.provider.credential,
            algod_address=str(self.provider.algod_url),
        )



"""Annotated types for Pydantic models."""

from functools import partial
from typing import Annotated, Union

from algosdk.constants import HASH_LEN, MAX_ASSET_DECIMALS
from annotated_types import Ge, Gt, Le, Len, Lt
from cytoolz import compose
from pydantic import AfterValidator, UrlConstraints
from pydantic_core import Url
from typing_extensions import TypeAliasType

from algobase.utils.url import decode_url_braces
from algobase.utils.validate import (
    validate_address,
    validate_arc3_sri,
    validate_base64,
    validate_contains_substring,
    validate_encoded_length,
    validate_hex,
    validate_is_power_of_10,
    validate_locale,
    validate_mime_type,
    validate_not_in,
    validate_not_ipfs_gateway,
    validate_type_compatibility,
)

# Generic types
Uint32 = Annotated[int, Ge(0), Lt(2**32)]
Uint64 = Annotated[int, Ge(0), Lt(2**64)]
Base64Str = Annotated[str, AfterValidator(validate_base64)]


# World Wide Web Consortium (W3C) types
MimeType = Annotated[
    str, AfterValidator(partial(validate_mime_type, primary_type=None))
]
ImageMimeType = Annotated[
    str, AfterValidator(partial(validate_mime_type, primary_type="image"))
]

# Unicode Common Locale Data Repository (CLDR) types
UnicodeLocale = Annotated[str, AfterValidator(validate_locale)]

# Algorand types
AlgorandHash = Annotated[bytes, Len(HASH_LEN, HASH_LEN)]  # 32 bytes
AlgorandAddress = Annotated[str, AfterValidator(validate_address)]

# Algorand Standard Asset (ASA) types
AsaDecimals = Annotated[Uint32, Ge(0), Le(MAX_ASSET_DECIMALS)]  # <= 19
AsaUnitName = Annotated[
    str, AfterValidator(partial(validate_encoded_length, max_length=8))
]
AsaAssetName = Annotated[
    str, AfterValidator(partial(validate_encoded_length, max_length=32))
]
AsaUrl = Annotated[
    str,
    AfterValidator(
        compose(
            partial(validate_encoded_length, max_length=96),
            decode_url_braces,
            partial(validate_type_compatibility, _type=Url),
        )
    ),
]
AsaFractionalNftTotal = Annotated[
    Uint64, Gt(1), AfterValidator(validate_is_power_of_10)
]

# Algorand ARC-16 types
Arc16Traits = dict[str, str | int]

# Algorand ARC-3 types
Arc3Url = Annotated[
    str,
    AfterValidator(
        compose(
            partial(validate_encoded_length, max_length=96),
            decode_url_braces,
            validate_not_ipfs_gateway,
            partial(
                validate_type_compatibility,
                _type=Annotated[Url, UrlConstraints(allowed_schemes=["https", "ipfs"])],
            ),
        )
    ),
]
Arc3LocalizedUrl = Annotated[
    str,
    AfterValidator(
        compose(
            partial(validate_encoded_length, max_length=96),
            partial(validate_contains_substring, substring="{locale}"),
            decode_url_braces,
            validate_not_ipfs_gateway,
            partial(
                validate_type_compatibility,
                _type=Annotated[Url, UrlConstraints(allowed_schemes=["https", "ipfs"])],
            ),
        )
    ),
]
Arc3Sri = Annotated[str, AfterValidator(validate_arc3_sri)]
Arc3Color = Annotated[str, Len(6, 6), AfterValidator(validate_hex)]
Arc3NonTraitProperties = TypeAliasType(
    "Arc3NonTraitProperties",
    Annotated[
        dict[
            str,
            Union[
                str,
                int,
                float,
                list[Union[str, int, float, "Arc3NonTraitProperties"]],  # type: ignore
                "Arc3NonTraitProperties",  # type: ignore
            ],
        ],
        AfterValidator(partial(validate_not_in, element="traits")),
    ],
)
